#include "Connect4.h"
#include <assert.h>


Connect4GameState::Connect4GameState()
{
	init();
}

Connect4GameState::~Connect4GameState()
{

}

Connect4GameState::Connect4GameState(const Connect4GameState &cpy)
{
	state.board = cpy.state.board;
	currentPlay = cpy.currentPlay;
}

void Connect4GameState::init()
{
	state.board.clear();
	state.board.resize(CONNECT4_BOARD_WIDTH);
	currentPlay = BLACK;
}

bool Connect4GameState::spawn(Connect4GameState &child, int move_idx)
{
	if(move_idx >= 0 && move_idx < CONNECT4_BOARD_WIDTH)
	{
		if(state.board[move_idx].size() < CONNECT4_BOARD_HEIGHT)
		{
			child.state.board = state.board;
			child.state.board[move_idx].push_back(currentPlay);
			if(currentPlay == RED)
				child.currentPlay = BLACK;
			else
				child.currentPlay = RED;
			return true;
		}
		else
			return false;
	}
	else
		return false;
}

bool Connect4GameState::spawn(std::vector<Connect4GameState> &children)
{
	if(endState())
		return false;
	children.clear();
	Connect4GameState tmpState;
	for(int i = 0; i < CONNECT4_BOARD_WIDTH; ++i)
	{
		if(spawn(tmpState, i))
			children.push_back(tmpState);
	}
	return true;
}

int Connect4GameState::endState()
{
	bool tie = true;
	for(size_t i = 0; i < state.board.size(); ++i)
	{
		if(state.board[i].size() != CONNECT4_BOARD_HEIGHT)
		{
			tie = false;
			break;
		}
	}
	if(tie)
		return -1;

	Connect4StateEvaluator tmp;
	if(tmp.searchQuad(state, currentPlay) > 0)
		return 1;
	else
		return false;
}

bool Connect4GameState::isOpening()
{
	int moveCount = 0;
	for(size_t i = 0; i < state.board.size(); ++i)
	{
		moveCount += state.board[i].size();
	}
	if(moveCount > 1)
		return false;
	else
		return true;
}

void Connect4GameState::print()
{
	for(int j = CONNECT4_BOARD_HEIGHT-1; j >= 0; --j)
	{
		std::cout << "\t| ";
		for(int i = 0; i < CONNECT4_BOARD_WIDTH; ++i)
		{
			if(j < (int)state.board[i].size())
				std::cout << state.board[i][j] << " ";
			else
				std::cout << (char)EMPTY << " ";
		}
		std::cout << "|" << std::endl;
	}
	std::cout << "\t|-";
	for(int i = 0; i < CONNECT4_BOARD_WIDTH; ++i)
	{
		std::cout << "--";
	}
	std::cout << "|" << std::endl;
	std::cout << "\t  ";
	for(int i = 0; i < CONNECT4_BOARD_WIDTH; ++i)
	{
		std::cout << i+1 << " ";
	}
	std::cout << std::endl;

}

Connect4GameState& Connect4GameState::operator=(const Connect4GameState &rhs)
{
	state = rhs.state;
	currentPlay = rhs.currentPlay;
	return *this;
}

void Connect4StateEvaluator::setWeights(Connect4GameStateFeature &_weights)
{
	assert(_weights.feature.size() == CONNECT4_FEATURE_LENGTH);
	weights = _weights;
}

double Connect4StateEvaluator::funct(Connect4GameState &state)
{
	assert(weights.feature.size() == CONNECT4_FEATURE_LENGTH);

	Connect4Board board = state.getBoard();
	Connect4GameStateFeature descriptor;
	descriptor.feature[Connect4GameStateFeature::BLACK_DOUBLE] = searchDouble(board, BLACK);
	descriptor.feature[Connect4GameStateFeature::RED_DOUBLE] = searchDouble(board, RED);
	descriptor.feature[Connect4GameStateFeature::BLACK_TRIPLE] = searchTriple(board, BLACK);
	descriptor.feature[Connect4GameStateFeature::RED_TRIPLE] = searchTriple(board, RED);
	descriptor.feature[Connect4GameStateFeature::BLACK_QUAD] = searchQuad(board, BLACK);
	descriptor.feature[Connect4GameStateFeature::RED_QUAD] = searchQuad(board, RED);
	descriptor.feature[Connect4GameStateFeature::BLACK_CENTER] = avgDistanceFromCenter(board, BLACK);
	descriptor.feature[Connect4GameStateFeature::RED_CENTER] = avgDistanceFromCenter(board, RED);

	double value = 0;
	for(size_t i = 0; i < descriptor.feature.size(); ++i)
		value += descriptor.feature[i] * weights.feature[i];

	return value;
}


int Connect4StateEvaluator::searchK(const Connect4Board &state, char target, int searchfor)
{
	if(searchfor != 2 && searchfor != 3 && searchfor != 4)
		return 0;

	int cnt = 0;
	int num;
	for(size_t i = 0; i < state.board.size(); ++i)
	{
		for(size_t j = 0; j < state.board[i].size(); ++j)
		{
			if(state.board[i][j] != target)
				continue;

			//search right
			num = searchDirection(state, target, i, j, 1, 0);
			if(num == searchfor-1)
				cnt++;

			//search up right
			num = searchDirection(state, target, i, j, 1, 1);
			if(num == searchfor-1)
				cnt++;

			//search up
			num = searchDirection(state, target, i, j, 0, 1);
			if(num == searchfor-1)
				cnt++;

			//search up left
			num = searchDirection(state, target, i, j, -1, 1);
			if(num == searchfor-1)
				cnt++;

			//search left
			num = searchDirection(state, target, i, j, -1, 0);
			if(num == searchfor-1)
				cnt++;

			//I don't believe these directions need to be searched, this is redundant
			////search down left
			//num = searchDirection(state, target, i, j, -1, -1);
			//if(num == searchfor-1)
			//	cnt++;

			////search down
			//num = searchDirection(state, target, i, j, 0, -1);
			//if(num == searchfor-1)
			//	cnt++;

			////search down right
			//num = searchDirection(state, target, i, j, 1, -1);
			//if(num == searchfor-1)
			//	cnt++;
		}
	}
	return cnt;
}

int Connect4StateEvaluator::searchDouble(const Connect4Board &state, char target)
{
	return searchK(state, target, 2);
}

int Connect4StateEvaluator::searchTriple(const Connect4Board &state, char target)
{
	return searchK(state, target, 3);
}

int Connect4StateEvaluator::searchQuad(const Connect4Board &state, char target)
{
	return searchK(state, target, 4);
}

int Connect4StateEvaluator::searchDirection(const Connect4Board &state, char target, int init_x, int init_y, int incr_x, int incr_y)
{
	int neighborCount = 0;
	int i = 1;
	while(i <= MAX_NEIGHBORS) //search all neighbors in 1 direction
	{
		init_x += incr_x;
		init_y += incr_y;
		if(init_x >= (int)state.board.size() || init_x < 0) //outside of board
		{
			neighborCount = 0;
			break;
		}
		if(init_y >= (int)state.board[init_x].size() || init_y < 0) // empty or outside
		{
			++i;
			continue;
		}
		else
		{
			if(state.board[init_x][init_y] != target) //found an opposite token in path
			{
				neighborCount = 0;
				break;
			}
			else
				neighborCount++;
		}
		++i;	
	}
	return neighborCount;
}

double Connect4StateEvaluator::avgDistanceFromCenter(const Connect4Board &state, char target)
{
	double center = (double)state.board.size()/2.0;
	double avgDist = 0;
	int avgCount = 0;
	for(size_t i = 0; i < state.board.size(); ++i)
	{
		for(size_t j = 0; j < state.board[i].size(); ++j)
		{
			if(state.board[i][j] == target)
			{
				avgDist += center - fabs((double)i - center);
				avgCount++;
			}
		}
	}
	avgDist /= (double)avgCount;

	return avgDist;
}

Connect4Game::Connect4Game()
{
	currentState.init();
	players.clear();
}

Connect4Game::~Connect4Game()
{
	players.clear();
}

Connect4Game::Connect4Game(const Connect4Game &cpy)
{
	currentState = cpy.currentState;
	players = cpy.players;
}

Connect4Game::Connect4Game(std::vector<game::Player<Connect4GameState, Connect4StateEvaluator>*> _players)
{
	setPlayers(_players);
	currentState.init();
}

void Connect4Game::setPlayers(std::vector<game::Player<Connect4GameState, Connect4StateEvaluator>*> _players)
{
	assert(_players.size() == 2);
	players = _players;
};

int Connect4Game::play()
{
	int currentPlayer = 0;
	while(1)
	{
		std::cout << std::endl;
		currentState.print();
		std::cout << std::endl;
		players[currentPlayer]->playTurn(currentState);
		std::cout << std::endl;
		if(currentPlayer)
			currentState.currentPlay = RED;
		else
			currentState.currentPlay = BLACK;

		int end = currentState.endState();
		if(end > 0)
		{
			if(currentPlayer)
				std::cout << "Red Player Wins!\n\n" << std::endl;
			else
				std::cout << "Black Player Wins!\n\n" << std::endl;
			break;
		}
		else if(end < 0)
		{
			std::cout << "It's a tie!!!\n\n" << std::endl;
			break;
		}

		if(currentPlayer)
		{
			currentPlayer = 0;
			currentState.currentPlay = BLACK;
			std::cout << "Black Player Turn: " << std::endl;
		}
		else
		{
			currentPlayer = 1;
			currentState.currentPlay = RED;
			std::cout << "Red Player Turn: " << std::endl;
		}
	}

	currentState.print();
	std::cout << "\n\n\tGAME OVER!\n" << std::endl;

	return currentPlayer;
}